# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common import extensions
from trove.extensions.common import service as common_service
from trove.extensions.mysql import service as mysql_service


class Mysql(extensions.ExtensionDescriptor):

    def get_name(self):
        return "Mysql"

    def get_description(self):
        return "Non essential MySQL services such as users and schemas"

    def get_alias(self):
        return "MYSQL"

    def get_namespace(self):
        return "http://TBD"

    def get_updated(self):
        return "2011-01-22T13:25:27-06:00"

    def get_resources(self):
        resources = []

        resource = extensions.ResourceExtension(
            'databases',
            mysql_service.SchemaController(),
            parent={'member_name': 'instance',
                    'collection_name': '{tenant_id}/instances'})
        resources.append(resource)

        resource = extensions.ResourceExtension(
            'users',
            mysql_service.UserController(),
            parent={'member_name': 'instance',
                    'collection_name': '{tenant_id}/instances'},
            member_actions={'update': 'PUT'},
            collection_actions={'update_all': 'PUT'})
        resources.append(resource)

        collection_url = '{tenant_id}/instances/:instance_id/users'
        resource = extensions.ResourceExtension(
            'databases',
            mysql_service.UserAccessController(),
            parent={'member_name': 'user',
                    'collection_name': collection_url},
            collection_actions={'update': 'PUT'})
        resources.append(resource)

        resource = extensions.ResourceExtension(
            'root',
            common_service.RootController(),
            parent={'member_name': 'instance',
                    'collection_name': '{tenant_id}/instances'},
            collection_actions={'delete': 'DELETE'})
        resources.append(resource)

        resource = extensions.ResourceExtension(
            'root',
            common_service.RootController(),
            parent={'member_name': 'instance',
                    'collection_name': '{tenant_id}/clusters'})
        resources.append(resource)

        return resources



# Copyright 2015 Tesora Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from swiftclient.client import ClientException
import tempfile

from troveclient.compat import exceptions

from trove.common import cfg
from trove.guestagent.common import operating_system
from trove.guestagent import guest_log
from trove.tests.config import CONFIG
from trove.tests.scenario.helpers.test_helper import DataType
from trove.tests.scenario import runners
from trove.tests.scenario.runners.test_runners import SkipKnownBug
from trove.tests.scenario.runners.test_runners import TestRunner


CONF = cfg.CONF


class GuestLogRunner(TestRunner):

    def __init__(self):
        super(GuestLogRunner, self).__init__()

        self.container = CONF.guest_log_container_name
        self.prefix_pattern = '%(instance_id)s/%(datastore)s-%(log)s/'
        self.stopped_log_details = None
        self._last_log_published = {}
        self._last_log_contents = {}

    def _get_last_log_published(self, log_name):
        return self._last_log_published.get(log_name, None)

    def _set_last_log_published(self, log_name, published):
        self._last_log_published[log_name] = published

    def _get_last_log_contents(self, log_name):
        return self._last_log_contents.get(log_name, [])

    def _set_last_log_contents(self, log_name, published):
        self._last_log_contents[log_name] = published

    def _get_exposed_user_log_names(self):
        """Returns the full list of exposed user logs."""
        return self.test_helper.get_exposed_user_log_names()

    def _get_exposed_user_log_name(self):
        """Return the first exposed user log name."""
        return self.test_helper.get_exposed_user_log_names()[0]

    def _get_unexposed_sys_log_name(self):
        """Return the first unexposed sys log name."""
        return self.test_helper.get_unexposed_sys_log_names()[0]

    def run_test_log_list(self):
        self.assert_log_list(self.auth_client,
                             self.test_helper.get_exposed_log_list())

    def assert_log_list(self, client, expected_list):
        log_list = list(client.instances.log_list(self.instance_info.id))
        log_names = list(ll.name for ll in log_list)
        self.assert_list_elements_equal(expected_list, log_names)
        self.register_debug_inst_ids(self.instance_info.id)

    def run_test_admin_log_list(self):
        self.assert_log_list(self.admin_client,
                             self.test_helper.get_full_log_list())

    def run_test_log_show(self):
        log_pending = self._set_zero_or_none()
        log_name = self._get_exposed_user_log_name()
        self.assert_log_show(self.auth_client,
                             log_name,
                             expected_published=0,
                             expected_pending=log_pending)

    def _set_zero_or_none(self):
        """This attempts to handle the case where an existing instance
        is used.  Values that would normally be '0' are not, and must
        be ignored.
        """
        value = 0
        if self.is_using_existing_instance:
            value = None
        return value

    def assert_log_show(self, client, log_name,
                        expected_http_code=200,
                        expected_type=guest_log.LogType.USER.name,
                        expected_status=guest_log.LogStatus.Disabled.name,
                        expected_published=None, expected_pending=None,
                        is_admin=False):
        self.report.log("Executing log_show for log '%s'" % log_name)
        log_details = client.instances.log_show(
            self.instance_info.id, log_name)
        self.assert_client_code(client, expected_http_code)
        self.assert_log_details(
            log_details, log_name,
            expected_type=expected_type,
            expected_status=expected_status,
            expected_published=expected_published,
            expected_pending=expected_pending,
            is_admin=is_admin)

    def assert_log_details(self, log_details, expected_log_name,
                           expected_type=guest_log.LogType.USER.name,
                           expected_status=guest_log.LogStatus.Disabled.name,
                           expected_published=None, expected_pending=None,
                           is_admin=False):
        """Check that the action generates the proper response data.
        For log_published and log_pending, setting the value to 'None'
        will skip that check (useful when using an existing instance,
        as there may be pending things in user logs right from the get-go)
        and setting it to a value other than '0' will verify that the actual
        value is '>=value' (since it's impossible to know what the actual
        value will be at any given time). '0' will still match exclusively.
        """
        self.report.log("Validating log details for log '%s'" %
                        expected_log_name)
        self._set_last_log_published(expected_log_name, log_details.published)
        self.assert_equal(expected_log_name, log_details.name,
                          "Wrong log name for '%s' log" % expected_log_name)
        self.assert_equal(expected_type, log_details.type,
                          "Wrong log type for '%s' log" % expected_log_name)
        current_status = log_details.status.replace(' ', '_')
        if not isinstance(expected_status, list):
            expected_status = [expected_status]
        self.assert_is_sublist([current_status], expected_status,
                               "Wrong log status for '%s' log" %
                               expected_log_name)
        if expected_published is None:
            pass
        elif expected_published == 0:
            self.assert_equal(0, log_details.published,
                              "Wrong log published for '%s' log" %
                              expected_log_name)
        else:
            self.assert_true(log_details.published >= expected_published,
                             "Missing log published for '%s' log: "
                             "expected %d, got %d" %
                             (expected_log_name, expected_published,
                              log_details.published))
        if expected_pending is None:
            pass
        elif expected_pending == 0:
            self.assert_equal(0, log_details.pending,
                              "Wrong log pending for '%s' log" %
                              expected_log_name)
        else:
            self.assert_true(log_details.pending >= expected_pending,
                             "Missing log pending for '%s' log: "
                             "expected %d, got %d" %
                             (expected_log_name, expected_pending,
                              log_details.pending))

        container = self.container
        prefix = self.prefix_pattern % {
            'instance_id': self.instance_info.id,
            'datastore': CONFIG.dbaas_datastore,
            'log': expected_log_name}
        metafile = prefix.rstrip('/') + '_metafile'

        if expected_published == 0:
            self.assert_storage_gone(container, prefix, metafile,
                                     is_admin=is_admin)
            container = 'None'
            prefix = 'None'
        else:
            self.assert_storage_exists(container, prefix, metafile,
                                       is_admin=is_admin)

        self.assert_equal(container, log_details.container,
                          "Wrong log container for '%s' log" %
                          expected_log_name)
        self.assert_equal(prefix, log_details.prefix,
                          "Wrong log prefix for '%s' log" % expected_log_name)
        self.assert_equal(metafile, log_details.metafile,
                          "Wrong log metafile for '%s' log" %
                          expected_log_name)

    def assert_log_enable(self, client, log_name,
                          expected_http_code=200,
                          expected_type=guest_log.LogType.USER.name,
                          expected_status=guest_log.LogStatus.Disabled.name,
                          expected_published=None, expected_pending=None):
        self.report.log("Executing log_enable for log '%s'" % log_name)
        log_details = client.instances.log_action(
            self.instance_info.id, log_name, enable=True)
        self.assert_client_code(client, expected_http_code)
        self.assert_log_details(
            log_details, log_name,
            expected_type=expected_type,
            expected_status=expected_status,
            expected_published=expected_published,
            expected_pending=expected_pending)

    def assert_log_disable(self, client, log_name, discard=None,
                           expected_http_code=200,
                           expected_type=guest_log.LogType.USER.name,
                           expected_status=guest_log.LogStatus.Disabled.name,
                           expected_published=None, expected_pending=None):
        self.report.log("Executing log_disable for log '%s' (discard: %s)" %
                        (log_name, discard))
        log_details = client.instances.log_action(
            self.instance_info.id, log_name, disable=True, discard=discard)
        self.assert_client_code(client, expected_http_code)
        self.assert_log_details(
            log_details, log_name,
            expected_type=expected_type,
            expected_status=expected_status,
            expected_published=expected_published,
            expected_pending=expected_pending)

    def assert_log_publish(self, client, log_name, disable=None, discard=None,
                           expected_http_code=200,
                           expected_type=guest_log.LogType.USER.name,
                           expected_status=guest_log.LogStatus.Disabled.name,
                           expected_published=None, expected_pending=None,
                           is_admin=False):
        self.report.log("Executing log_publish for log '%s' (disable: %s  "
                        "discard: %s)" %
                        (log_name, disable, discard))
        log_details = client.instances.log_action(
            self.instance_info.id, log_name, publish=True, disable=disable,
            discard=discard)
        self.assert_client_code(client, expected_http_code)
        self.assert_log_details(
            log_details, log_name,
            expected_type=expected_type,
            expected_status=expected_status,
            expected_published=expected_published,
            expected_pending=expected_pending,
            is_admin=is_admin)

    def assert_log_discard(self, client, log_name,
                           expected_http_code=200,
                           expected_type=guest_log.LogType.USER.name,
                           expected_status=guest_log.LogStatus.Disabled.name,
                           expected_published=None, expected_pending=None):
        self.report.log("Executing log_discard for log '%s'" % log_name)
        log_details = client.instances.log_action(
            self.instance_info.id, log_name, discard=True)
        self.assert_client_code(client, expected_http_code)
        self.assert_log_details(
            log_details, log_name,
            expected_type=expected_type,
            expected_status=expected_status,
            expected_published=expected_published,
            expected_pending=expected_pending)

    def assert_storage_gone(self, container, prefix, metafile, is_admin=False):
        if is_admin:
            swift_client = self.admin_swift_client
        else:
            swift_client = self.swift_client

        try:
            headers, container_files = swift_client.get_container(
                container, prefix=prefix)
            self.assert_equal(0, len(container_files),
                              "Found files in %s/%s: %s" %
                              (container, prefix, container_files))
        except ClientException as ex:
            if ex.http_status == 404:
                self.report.log("Container '%s' does not exist" %
                                container)
                pass
            else:
                raise
        try:
            swift_client.get_object(container, metafile)
            self.fail("Found metafile after discard: %s" % metafile)
        except ClientException as ex:
            if ex.http_status == 404:
                self.report.log("Metafile '%s' gone as expected" %
                                metafile)
                pass
            else:
                raise

    def assert_storage_exists(self, container, prefix, metafile,
                              is_admin=False):
        if is_admin:
            swift_client = self.admin_swift_client
        else:
            swift_client = self.swift_client

        try:
            headers, container_files = swift_client.get_container(
                container, prefix=prefix)
            self.assert_true(len(container_files) > 0,
                             "No files found in %s/%s" %
                             (container, prefix))
        except ClientException as ex:
            if ex.http_status == 404:
                self.fail("Container '%s' does not exist" % container)
            else:
                raise
        try:
            swift_client.get_object(container, metafile)
        except ClientException as ex:
            if ex.http_status == 404:
                self.fail("Missing metafile: %s" % metafile)
            else:
                raise

    def run_test_log_enable_sys(self,
                                expected_exception=exceptions.BadRequest,
                                expected_http_code=400):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_enable_fails(
            self.admin_client,
            expected_exception, expected_http_code,
            log_name)

    def assert_log_enable_fails(self, client,
                                expected_exception, expected_http_code,
                                log_name):
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.instances.log_action,
                           self.instance_info.id, log_name, enable=True)

    def run_test_log_disable_sys(self,
                                 expected_exception=exceptions.BadRequest,
                                 expected_http_code=400):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_disable_fails(
            self.admin_client,
            expected_exception, expected_http_code,
            log_name)

    def assert_log_disable_fails(self, client,
                                 expected_exception, expected_http_code,
                                 log_name, discard=None):
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.instances.log_action,
                           self.instance_info.id, log_name, disable=True,
                           discard=discard)

    def run_test_log_show_unauth_user(self,
                                      expected_exception=exceptions.NotFound,
                                      expected_http_code=404):
        log_name = self._get_exposed_user_log_name()
        self.assert_log_show_fails(
            self.unauth_client,
            expected_exception, expected_http_code,
            log_name)

    def assert_log_show_fails(self, client,
                              expected_exception, expected_http_code,
                              log_name):
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.instances.log_show,
                           self.instance_info.id, log_name)

    def run_test_log_list_unauth_user(self,
                                      expected_exception=exceptions.NotFound,
                                      expected_http_code=404):
        client = self.unauth_client
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.instances.log_list,
                           self.instance_info.id)

    def run_test_log_generator_unauth_user(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        log_name = self._get_exposed_user_log_name()
        self.assert_log_generator_unauth_user(
            self.unauth_client, log_name,
            expected_exception, expected_http_code)

    def assert_log_generator_unauth_user(self, client, log_name,
                                         expected_exception,
                                         expected_http_code,
                                         publish=None):
        raise SkipKnownBug(runners.BUG_UNAUTH_TEST_WRONG)
        # self.assert_raises(expected_exception, expected_http_code,
        #                    client, client.instances.log_generator,
        #                    self.instance_info.id, log_name, publish=publish)

    def run_test_log_generator_publish_unauth_user(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        log_name = self._get_exposed_user_log_name()
        self.assert_log_generator_unauth_user(
            self.unauth_client, log_name,
            expected_exception, expected_http_code,
            publish=True)

    def run_test_log_show_unexposed_user(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_show_fails(
            self.auth_client,
            expected_exception, expected_http_code,
            log_name)

    def run_test_log_enable_unexposed_user(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_enable_fails(
            self.auth_client,
            expected_exception, expected_http_code,
            log_name)

    def run_test_log_disable_unexposed_user(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_disable_fails(
            self.auth_client,
            expected_exception, expected_http_code,
            log_name)

    def run_test_log_publish_unexposed_user(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_publish_fails(
            self.auth_client,
            expected_exception, expected_http_code,
            log_name)

    def assert_log_publish_fails(self, client,
                                 expected_exception, expected_http_code,
                                 log_name,
                                 disable=None, discard=None):
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.instances.log_action,
                           self.instance_info.id, log_name, publish=True,
                           disable=disable, discard=discard)

    def run_test_log_discard_unexposed_user(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_discard_fails(
            self.auth_client,
            expected_exception, expected_http_code,
            log_name)

    def assert_log_discard_fails(self, client,
                                 expected_exception, expected_http_code,
                                 log_name):
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.instances.log_action,
                           self.instance_info.id, log_name, discard=True)

    def run_test_log_enable_user(self):
        expected_status = guest_log.LogStatus.Ready.name
        expected_pending = 1
        if self.test_helper.log_enable_requires_restart():
            expected_status = guest_log.LogStatus.Restart_Required.name
            # if using an existing instance, there may already be something
            expected_pending = self._set_zero_or_none()

        for log_name in self._get_exposed_user_log_names():
            self.assert_log_enable(
                self.auth_client,
                log_name,
                expected_status=expected_status,
                expected_published=0, expected_pending=expected_pending)

    def run_test_log_enable_flip_user(self):
        # for restart required datastores, test that flipping them
        # back to disabled returns the status to 'Disabled'
        # from 'Restart_Required'
        if self.test_helper.log_enable_requires_restart():
            # if using an existing instance, there may already be something
            expected_pending = self._set_zero_or_none()

            for log_name in self._get_exposed_user_log_names():
                self.assert_log_disable(
                    self.auth_client,
                    log_name,
                    expected_status=guest_log.LogStatus.Disabled.name,
                    expected_published=0, expected_pending=expected_pending)
                self.assert_log_enable(
                    self.auth_client,
                    log_name,
                    expected_status=guest_log.LogStatus.Restart_Required.name,
                    expected_published=0, expected_pending=expected_pending)

    def run_test_restart_datastore(self, expected_http_code=202):
        if self.test_helper.log_enable_requires_restart():
            instance_id = self.instance_info.id
            # we need to wait until the heartbeat flips the instance
            # back into 'ACTIVE' before we issue the restart command
            expected_states = ['RESTART_REQUIRED', 'HEALTHY']
            self.assert_instance_action(instance_id, expected_states)
            client = self.auth_client
            client.instances.restart(instance_id)
            self.assert_client_code(client, expected_http_code)

    def run_test_wait_for_restart(self, expected_states=['REBOOT', 'HEALTHY']):
        if self.test_helper.log_enable_requires_restart():
            self.assert_instance_action(self.instance_info.id, expected_states)

    def run_test_log_publish_user(self):
        for log_name in self._get_exposed_user_log_names():
            self.assert_log_publish(
                self.auth_client,
                log_name,
                expected_status=[guest_log.LogStatus.Published.name,
                                 guest_log.LogStatus.Partial.name],
                expected_published=1, expected_pending=None)

    def run_test_add_data(self):
        self.test_helper.add_data(DataType.micro, self.get_instance_host())

    def run_test_verify_data(self):
        self.test_helper.verify_data(DataType.micro, self.get_instance_host())

    def run_test_log_publish_again_user(self):
        for log_name in self._get_exposed_user_log_names():
            self.assert_log_publish(
                self.auth_client,
                log_name,
                expected_status=[guest_log.LogStatus.Published.name,
                                 guest_log.LogStatus.Partial.name],
                expected_published=self._get_last_log_published(log_name),
                expected_pending=None)

    def run_test_log_generator_user(self):
        for log_name in self._get_exposed_user_log_names():
            self.assert_log_generator(
                self.auth_client,
                log_name,
                lines=2, expected_lines=2)

    def assert_log_generator(self, client, log_name, publish=False,
                             lines=4, expected_lines=None,
                             swift_client=None):
        self.report.log("Executing log_generator for log '%s' (publish: %s)" %
                        (log_name, publish))

        if publish:
            client.instances.log_action(self.instance_info.id, log_name,
                                        publish=True)

        log_gen = client.instances.log_generator(
            self.instance_info.id, log_name,
            lines=lines, swift=swift_client)
        log_contents = "".join([chunk for chunk in log_gen()])

        self.report.log("Returned %d lines for log '%s': %s" % (
            len(log_contents.splitlines()), log_name, log_contents))

        self._set_last_log_contents(log_name, log_contents)
        if expected_lines:
            self.assert_equal(expected_lines,
                              len(log_contents.splitlines()),
                              "Wrong line count for '%s' log" % log_name)
        else:
            self.assert_true(len(log_contents.splitlines()) <= lines,
                             "More than %d lines found for '%s' log" %
                             (lines, log_name))

    def run_test_log_generator_publish_user(self):
        for log_name in self._get_exposed_user_log_names():
            self.assert_log_generator(
                self.auth_client,
                log_name, publish=True,
                lines=3, expected_lines=3)

    def run_test_log_generator_swift_client_user(self):
        swift_client = self.swift_client
        for log_name in self._get_exposed_user_log_names():
            self.assert_log_generator(
                self.auth_client,
                log_name, publish=True,
                lines=3, expected_lines=3,
                swift_client=swift_client)

    def run_test_add_data_again(self):
        # Add some more data so we have at least 3 log data files
        self.test_helper.add_data(DataType.micro2, self.get_instance_host())

    def run_test_verify_data_again(self):
        self.test_helper.verify_data(DataType.micro2, self.get_instance_host())

    def run_test_log_generator_user_by_row(self):
        log_name = self._get_exposed_user_log_name()
        self.assert_log_publish(
            self.auth_client,
            log_name,
            expected_status=[guest_log.LogStatus.Published.name,
                             guest_log.LogStatus.Partial.name],
            expected_published=self._get_last_log_published(log_name),
            expected_pending=None)
        # Now get the full contents of the log
        self.assert_log_generator(self.auth_client, log_name, lines=100000)
        log_lines = len(self._get_last_log_contents(log_name).splitlines())
        # cap at 100, so the test can't run away if something goes wrong
        log_lines = min(log_lines, 100)
        # Make sure we get the right number of log lines back each time
        for lines in range(1, log_lines):
            self.assert_log_generator(
                self.auth_client,
                log_name, lines=lines, expected_lines=lines)

    def run_test_log_save_user(self):
        for log_name in self._get_exposed_user_log_names():
            self.assert_test_log_save(self.auth_client, log_name)

    def run_test_log_save_publish_user(self):
        for log_name in self._get_exposed_user_log_names():
            self.assert_test_log_save(self.auth_client, log_name, publish=True)

    def assert_test_log_save(self, client, log_name, publish=False):
        # generate the file
        self.report.log("Executing log_save for log '%s' (publish: %s)" %
                        (log_name, publish))

        if publish:
            client.instances.log_action(self.instance_info.id,
                                        log_name=log_name,
                                        publish=True)

        with tempfile.NamedTemporaryFile() as temp_file:
            client.instances.log_save(self.instance_info.id,
                                      log_name=log_name,
                                      filename=temp_file.name)
            file_contents = operating_system.read_file(temp_file.name)
            # now grab the contents ourselves
            self.assert_log_generator(client, log_name, lines=100000)
            # and compare them
            self.assert_equal(self._get_last_log_contents(log_name),
                              file_contents)

    def run_test_log_discard_user(self):
        for log_name in self._get_exposed_user_log_names():
            self.assert_log_discard(
                self.auth_client,
                log_name,
                expected_status=guest_log.LogStatus.Ready.name,
                expected_published=0, expected_pending=1)

    def run_test_log_disable_user(self):
        expected_status = guest_log.LogStatus.Disabled.name
        if self.test_helper.log_enable_requires_restart():
            expected_status = guest_log.LogStatus.Restart_Required.name
        for log_name in self._get_exposed_user_log_names():
            self.assert_log_disable(
                self.auth_client,
                log_name,
                expected_status=expected_status,
                expected_published=0, expected_pending=1)

    def run_test_log_show_after_stop_details(self):
        log_name = self._get_exposed_user_log_name()
        self.stopped_log_details = self.auth_client.instances.log_show(
            self.instance_info.id, log_name)
        self.assert_is_not_none(self.stopped_log_details)

    def run_test_add_data_again_after_stop(self):
        # Add some more data to make sure logging has stopped
        self.test_helper.add_data(DataType.micro3, self.get_instance_host())

    def run_test_verify_data_again_after_stop(self):
        self.test_helper.verify_data(DataType.micro3, self.get_instance_host())

    def run_test_log_show_after_stop(self):
        log_name = self._get_exposed_user_log_name()
        self.assert_log_show(
            self.auth_client, log_name,
            expected_published=self.stopped_log_details.published,
            expected_pending=self.stopped_log_details.pending)

    def run_test_log_enable_user_after_stop(self):
        expected_status = guest_log.LogStatus.Ready.name
        expected_pending = 1
        if self.test_helper.log_enable_requires_restart():
            expected_status = guest_log.LogStatus.Restart_Required.name

        log_name = self._get_exposed_user_log_name()
        self.assert_log_enable(
            self.auth_client,
            log_name,
            expected_status=expected_status,
            expected_published=0, expected_pending=expected_pending)

    def run_test_add_data_again_after_stop_start(self):
        # Add some more data to make sure logging has started again
        self.test_helper.add_data(DataType.micro4, self.get_instance_host())

    def run_test_verify_data_again_after_stop_start(self):
        self.test_helper.verify_data(DataType.micro4, self.get_instance_host())

    def run_test_log_publish_after_stop_start(self):
        log_name = self._get_exposed_user_log_name()
        self.assert_log_publish(
            self.auth_client,
            log_name,
            expected_status=[guest_log.LogStatus.Published.name,
                             guest_log.LogStatus.Partial.name],
            expected_published=self._get_last_log_published(log_name) + 1,
            expected_pending=None)

    def run_test_log_disable_user_after_stop_start(self):
        expected_status = guest_log.LogStatus.Disabled.name
        if self.test_helper.log_enable_requires_restart():
            expected_status = guest_log.LogStatus.Restart_Required.name
        log_name = self._get_exposed_user_log_name()
        self.assert_log_disable(
            self.auth_client,
            log_name, discard=True,
            expected_status=expected_status,
            expected_published=0, expected_pending=1)

    def run_test_log_show_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_show(
            self.admin_client,
            log_name,
            expected_type=guest_log.LogType.SYS.name,
            expected_status=[guest_log.LogStatus.Ready.name,
                             guest_log.LogStatus.Partial.name],
            expected_published=0, expected_pending=1,
            is_admin=True
        )

    def run_test_log_publish_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_publish(
            self.admin_client,
            log_name,
            expected_type=guest_log.LogType.SYS.name,
            expected_status=[guest_log.LogStatus.Partial.name,
                             guest_log.LogStatus.Published.name],
            expected_published=1, expected_pending=None,
            is_admin=True)

    def run_test_log_publish_again_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_publish(
            self.admin_client,
            log_name,
            expected_type=guest_log.LogType.SYS.name,
            expected_status=[guest_log.LogStatus.Partial.name,
                             guest_log.LogStatus.Published.name],
            expected_published=self._get_last_log_published(log_name) + 1,
            expected_pending=None,
            is_admin=True)

    def run_test_log_generator_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_generator(
            self.admin_client,
            log_name,
            lines=4, expected_lines=4)

    def run_test_log_generator_publish_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_generator(
            self.admin_client,
            log_name, publish=True,
            lines=4, expected_lines=4)

    def run_test_log_generator_swift_client_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_generator(
            self.admin_client,
            log_name, publish=True,
            lines=4, expected_lines=4,
            swift_client=self.admin_swift_client)

    def run_test_log_save_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_test_log_save(
            self.admin_client,
            log_name)

    def run_test_log_save_publish_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_test_log_save(
            self.admin_client,
            log_name,
            publish=True)

    def run_test_log_discard_sys(self):
        log_name = self._get_unexposed_sys_log_name()
        self.assert_log_discard(
            self.admin_client,
            log_name,
            expected_type=guest_log.LogType.SYS.name,
            expected_status=guest_log.LogStatus.Ready.name,
            expected_published=0, expected_pending=1)


class CassandraGuestLogRunner(GuestLogRunner):

    def run_test_log_show(self):
        log_name = self._get_exposed_user_log_name()
        self.assert_log_show(self.auth_client,
                             log_name,
                             expected_published=0,
                             expected_pending=None)



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from trove.common.db import models


class PostgreSQLSchema(models.DatastoreSchema):
    """Represents a PostgreSQL schema and its associated properties."""

    name_regex = re.compile(str(r'^[\u0001-\u007F\u0080-\uFFFF]+[^\s]$'))

    def __init__(self, name=None, collate=None, character_set=None,
                 deserializing=False):
        super(PostgreSQLSchema, self).__init__(name=name,
                                               deserializing=deserializing)
        self.collate = collate
        self.character_set = character_set

    @property
    def collate(self):
        return self._collate

    @collate.setter
    def collate(self, value):
        self._collate = value

    @property
    def character_set(self):
        return self._character_set

    @character_set.setter
    def character_set(self, value):
        self._character_set = value

    @property
    def _max_schema_name_length(self):
        return 63

    def _is_valid_schema_name(self, value):
        return self.name_regex.match(value) is not None


class PostgreSQLUser(models.DatastoreUser):
    """Represents a PostgreSQL user and its associated properties."""

    root_username = 'postgres'

    @property
    def _max_user_name_length(self):
        return 63

    @property
    def schema_model(self):
        return PostgreSQLSchema



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
from oslo_utils import encodeutils
import stevedore

from trove.common import base_exception as exception
from trove.common import cfg
from trove.common.i18n import _

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class ModuleDriverManager(object):

    MODULE_DRIVER_NAMESPACE = 'trove.guestagent.module.drivers'

    def __init__(self):
        LOG.info('Initializing module driver manager.')

        self._drivers = {}
        self._module_types = [mt.lower() for mt in CONF.module_types]

        self._load_drivers()

    def _load_drivers(self):
        manager = stevedore.enabled.EnabledExtensionManager(
            namespace=self.MODULE_DRIVER_NAMESPACE,
            check_func=self._check_extension,
            invoke_on_load=True,
            invoke_kwds={})
        try:
            manager.map(self.add_driver_extension)
        except stevedore.exception.NoMatches:
            LOG.info("No module drivers loaded")

    def _check_extension(self, extension):
        """Checks for required methods in driver objects."""
        driver = extension.obj
        supported = False
        try:
            LOG.info('Loading Module driver: %s', driver.get_type())
            if driver.get_type() != driver.get_type().lower():
                raise AttributeError(_("Driver 'type' must be lower-case"))
            LOG.debug('  description: %s', driver.get_description())
            LOG.debug('  updated    : %s', driver.get_updated())
            required_attrs = ['apply', 'remove']
            for attr in required_attrs:
                if not hasattr(driver, attr):
                    raise AttributeError(
                        _("Driver '%(type)s' missing attribute: %(attr)s")
                        % {'type': driver.get_type(), 'attr': attr})
            if driver.get_type() in self._module_types:
                supported = True
            else:
                LOG.info("Driver '%s' not supported, skipping",
                         driver.get_type())
        except AttributeError as ex:
            LOG.exception("Exception loading module driver: %s",
                          encodeutils.exception_to_unicode(ex))

        return supported

    def add_driver_extension(self, extension):
        # Add a module driver from the extension.
        # If the stevedore manager is changed to one that doesn't
        # check the extension driver, then it should be done manually here
        # by calling self._check_extension(extension)
        driver = extension.obj
        driver_type = driver.get_type()
        LOG.info('Loaded module driver: %s', driver_type)

        if driver_type in self._drivers:
            raise exception.Error(_("Found duplicate driver: %s") %
                                  driver_type)
        self._drivers[driver_type] = driver

    def get_driver(self, driver_type):
        found = None
        if driver_type in self._drivers:
            found = self._drivers[driver_type]
        return found



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from trove.common import cfg
from trove.common.rpc import conductor_guest_serializer as gsz
from trove.common.rpc import conductor_host_serializer as hsz

from trove.tests.unittests import trove_testtools


CONF = cfg.CONF


class FakeInstance(object):
    def __init__(self):
        self.uuid = 'a3af1652-686a-4574-a916-2ef7e85136e5'

    @property
    def key(self):
        return 'mo79Y86Bp3bzQDWR31ihhVGfLBmeac'


class FakeContext(object):
    def __init__(self, instance_id=None, fields=None):
        self.instance_id = instance_id
        self.fields = fields


class TestConductorSerializer(trove_testtools.TestCase):

    def setUp(self):
        self.uuid = 'a3af1652-686a-4574-a916-2ef7e85136e5'
        self.key = 'mo79Y86Bp3bzQDWR31ihhVGfLBmeac'
        self.data = 'ELzWd81qtgcj2Gxc1ipbh0HgbvHGrgptDj3n4GNMBN0F2WtNdr'
        self.context = {'a': 'ij2J8AJLyz0rDqbjxy4jPVINhnK2jsBGpWRKIe3tUnUD',
                        'b': 32,
                        'c': {'a': 21, 'b': 22}}
        self.old_guest_id = gsz.CONF.guest_id
        gsz.CONF.guest_id = self.uuid
        super(TestConductorSerializer, self).setUp()

    def tearDown(self):
        gsz.CONF.guest_id = self.old_guest_id
        super(TestConductorSerializer, self).tearDown()

    def test_gsz_serialize_entity_nokey(self):
        sz = gsz.ConductorGuestSerializer(None, None)
        self.assertEqual(sz.serialize_entity(self.context, self.data),
                         self.data)

    def test_gsz_serialize_context_nokey(self):
        sz = gsz.ConductorGuestSerializer(None, None)
        self.assertEqual(sz.serialize_context(self.context),
                         self.context)

    @mock.patch('trove.common.rpc.conductor_host_serializer.'
                'get_instance_encryption_key',
                return_value='mo79Y86Bp3bzQDWR31ihhVGfLBmeac')
    def test_hsz_serialize_entity_nokey_noinstance(self, _):
        sz = hsz.ConductorHostSerializer(None, None)
        ctxt = FakeContext(instance_id=None)
        self.assertEqual(sz.serialize_entity(ctxt, self.data),
                         self.data)

    @mock.patch('trove.common.rpc.conductor_host_serializer.'
                'get_instance_encryption_key',
                return_value='mo79Y86Bp3bzQDWR31ihhVGfLBmeac')
    def test_hsz_serialize_context_nokey_noinstance(self, _):
        sz = hsz.ConductorHostSerializer(None, None)
        ctxt = FakeContext(instance_id=None)
        self.assertEqual(sz.serialize_context(ctxt), ctxt)

    @mock.patch('trove.common.rpc.conductor_host_serializer.'
                'get_instance_encryption_key',
                return_value='mo79Y86Bp3bzQDWR31ihhVGfLBmeac')
    def test_conductor_entity(self, _):
        guestsz = gsz.ConductorGuestSerializer(None, self.key)
        hostsz = hsz.ConductorHostSerializer(None, None)
        encrypted_entity = guestsz.serialize_entity(self.context, self.data)
        self.assertNotEqual(encrypted_entity, self.data)
        entity = hostsz.deserialize_entity(self.context, encrypted_entity)
        self.assertEqual(entity, self.data)

    @mock.patch('trove.common.rpc.conductor_host_serializer.'
                'get_instance_encryption_key',
                return_value='mo79Y86Bp3bzQDWR31ihhVGfLBmeac')
    def test_conductor_context(self, _):
        guestsz = gsz.ConductorGuestSerializer(None, self.key)
        hostsz = hsz.ConductorHostSerializer(None, None)
        encrypted_context = guestsz.serialize_context(self.context)
        self.assertNotEqual(encrypted_context, self.context)
        context = hostsz.deserialize_context(encrypted_context)
        self.assertEqual(context.get('instance_id'), self.uuid)
        context.pop('instance_id')
        self.assertDictEqual(context, self.context)



#    Copyright 2014 Rackspace Hosting
#    Copyright 2014 Hewlett-Packard Development Company, L.P.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import uuid

from unittest.mock import Mock
from unittest.mock import patch

from trove.backup import models as backup_models
from trove.common import cfg
from trove.common import clients
from trove.common import exception
from trove.common import neutron
from trove.datastore import models as datastore_models
from trove.instance import models
from trove.instance.models import DBInstance
from trove.instance.models import DBInstanceFault
from trove.instance.models import Instance
from trove.instance.models import InstanceServiceStatus
from trove.instance.models import SimpleInstance
from trove.instance.models import instance_encryption_key_cache
from trove.instance.service_status import ServiceStatuses
from trove.instance.tasks import InstanceTasks
from trove.taskmanager import api as task_api
from trove.tests.fakes import nova
from trove.tests.unittests import trove_testtools
from trove.tests.unittests.util import util

CONF = cfg.CONF


class SimpleInstanceTest(trove_testtools.TestCase):

    def setUp(self):
        super(SimpleInstanceTest, self).setUp()
        self.context = trove_testtools.TroveTestContext(self, is_admin=True)
        db_info = DBInstance(
            InstanceTasks.BUILDING, name="TestInstance")
        self.instance = SimpleInstance(
            None, db_info, InstanceServiceStatus(
                ServiceStatuses.BUILDING), ds_version=Mock(), ds=Mock(),
            locality='affinity')
        self.instance.context = self.context
        db_info.addresses = [{
            'type': 'private',
            'address': '123.123.123.123',
            'network': 'net-id-private'}, {
            'type': 'private',
            'address': '10.123.123.123',
            'network': 'net-id-private'}, {
            'type': 'public',
            'address': '15.123.123.123',
            'network': 'net-id-public'}]

        self.orig_ip_regex = CONF.ip_regex
        self.orig_black_list_regex = CONF.black_list_regex

    def tearDown(self):
        super(SimpleInstanceTest, self).tearDown()
        CONF.ip_start = None
        CONF.management_networks = []
        CONF.ip_regex = self.orig_ip_regex
        CONF.black_list_regex = self.orig_black_list_regex

        neutron.reset_management_networks()

    def test_get_root_on_create(self):
        root_on_create_val = Instance.get_root_on_create(
            'redis')
        self.assertFalse(root_on_create_val)

    def test_filter_ips_white_list(self):
        CONF.ip_regex = '^(15.|123.)'
        CONF.black_list_regex = '^10.123.123.*'
        ip = self.instance.get_visible_ip_addresses()
        self.assertEqual(2, len(ip))
        self.assertIn('123.123.123.123', ip[0].get('address'))
        self.assertIn('15.123.123.123', ip[1].get('address'))

    def test_filter_ips_black_list(self):
        CONF.ip_regex = '.*'
        CONF.black_list_regex = '^10.123.123.*'
        ip = self.instance.get_visible_ip_addresses()
        self.assertEqual(2, len(ip))
        self.assertNotIn('10.123.123.123', ip)

    def test_locality(self):
        self.assertEqual('affinity', self.instance.locality)

    def test_fault(self):
        fault_message = 'Error'
        fault_details = 'details'
        fault_date = 'now'
        temp_fault = Mock()
        temp_fault.message = fault_message
        temp_fault.details = fault_details
        temp_fault.updated = fault_date
        fault_mock = Mock(return_value=temp_fault)
        with patch.object(DBInstanceFault, 'find_by', fault_mock):
            fault = self.instance.fault
            self.assertEqual(fault_message, fault.message)
            self.assertEqual(fault_details, fault.details)
            self.assertEqual(fault_date, fault.updated)


class CreateInstanceTest(trove_testtools.TestCase):

    @patch.object(task_api.API, 'get_client', Mock(return_value=Mock()))
    def setUp(self):
        util.init_db()
        self.context = trove_testtools.TroveTestContext(self, is_admin=True)
        self.name = "name"
        self.flavor_id = 5
        self.image_id = "UUID"
        self.databases = []
        self.users = []
        self.datastore = datastore_models.DBDatastore.create(
            id=str(uuid.uuid4()),
            name='mysql' + str(uuid.uuid4()),
        )
        self.datastore_version = (
            datastore_models.DBDatastoreVersion.create(
                id=str(uuid.uuid4()),
                datastore_id=self.datastore.id,
                name="5.5" + str(uuid.uuid4()),
                manager="mysql",
                image_id="image_id",
                packages="",
                active=True))
        self.volume_size = 1
        self.az = "az"
        self.nics = None
        self.configuration = None
        self.tenant_id = "UUID"
        self.datastore_version_id = str(uuid.uuid4())

        self.db_info = DBInstance.create(
            name=self.name, flavor_id=self.flavor_id,
            tenant_id=self.tenant_id,
            volume_size=self.volume_size,
            datastore_version_id=self.datastore_version.id,
            task_status=InstanceTasks.BUILDING,
            configuration_id=self.configuration
        )

        self.backup_name = "name"
        self.descr = None
        self.backup_state = backup_models.BackupState.COMPLETED
        self.instance_id = self.db_info.id
        self.parent_id = None
        self.deleted = False

        self.backup = backup_models.DBBackup.create(
            name=self.backup_name,
            description=self.descr,
            tenant_id=self.tenant_id,
            state=self.backup_state,
            instance_id=self.instance_id,
            parent_id=self.parent_id,
            datastore_version_id=self.datastore_version.id,
            deleted=False
        )
        self.backup_id = self.backup.id

        self.orig_client = clients.create_nova_client
        clients.create_nova_client = nova.fake_create_nova_client

        self.orig_api = task_api.API(self.context).create_instance
        task_api.API(self.context).create_instance = Mock()
        self.run_with_quotas = models.run_with_quotas
        models.run_with_quotas = Mock()
        self.check = backup_models.DBBackup.check_swift_object_exist
        backup_models.DBBackup.check_swift_object_exist = Mock(
            return_value=True)
        self.locality = 'affinity'

        self.swift_verify_patch = patch.object(models.Backup,
                                               'verify_swift_auth_token')
        self.addCleanup(self.swift_verify_patch.stop)
        self.swift_verify_patch.start()

        super(CreateInstanceTest, self).setUp()

    @patch.object(task_api.API, 'get_client', Mock(return_value=Mock()))
    def tearDown(self):
        self.db_info.delete()
        self.backup.delete()
        self.datastore.delete()
        self.datastore_version.delete()
        clients.create_nova_client = self.orig_client
        task_api.API(self.context).create_instance = self.orig_api
        models.run_with_quotas = self.run_with_quotas
        backup_models.DBBackup.check_swift_object_exist = self.check
        self.backup.delete()
        self.db_info.delete()
        super(CreateInstanceTest, self).tearDown()

    def test_exception_on_invalid_backup_size(self):
        self.backup.size = 1.1
        self.backup.save()
        self.assertEqual(self.backup.id, self.backup_id)
        exc = self.assertRaises(
            exception.BackupTooLarge, models.Instance.create,
            self.context, self.name, self.flavor_id,
            self.image_id, self.databases, self.users,
            self.datastore, self.datastore_version,
            self.volume_size, self.backup_id,
            self.az, self.nics, self.configuration
        )
        self.assertIn("Backup is too large for "
                      "given flavor or volume.", str(exc))

    def test_can_restore_from_backup_with_almost_equal_size(self):
        # target size equals to "1Gb"
        self.backup.size = 0.99
        self.backup.save()
        instance = models.Instance.create(
            self.context, self.name, self.flavor_id,
            self.image_id, self.databases, self.users,
            self.datastore, self.datastore_version,
            self.volume_size, self.backup_id,
            self.az, self.nics, self.configuration)
        self.assertIsNotNone(instance)

    def test_can_instantiate_with_locality(self):
        # make sure the backup will fit
        self.backup.size = 0.2
        self.backup.save()
        instance = models.Instance.create(
            self.context, self.name, self.flavor_id,
            self.image_id, self.databases, self.users,
            self.datastore, self.datastore_version,
            self.volume_size, self.backup_id,
            self.az, self.nics, self.configuration,
            locality=self.locality)
        self.assertIsNotNone(instance)


class TestInstanceUpgrade(trove_testtools.TestCase):

    def setUp(self):
        self.context = trove_testtools.TroveTestContext(self, is_admin=True)
        util.init_db()

        self.datastore = datastore_models.DBDatastore.create(
            id=str(uuid.uuid4()),
            name='test' + str(uuid.uuid4()),
            default_version_id=str(uuid.uuid4()))

        self.datastore_version1 = datastore_models.DBDatastoreVersion.create(
            id=self.datastore.default_version_id,
            name='name' + str(uuid.uuid4()),
            image_id='old_image',
            packages=str(uuid.uuid4()),
            datastore_id=self.datastore.id,
            manager='test',
            active=1)

        self.datastore_version2 = datastore_models.DBDatastoreVersion.create(
            id=str(uuid.uuid4()),
            name='name' + str(uuid.uuid4()),
            image_id='new_image',
            packages=str(uuid.uuid4()),
            datastore_id=self.datastore.id,
            manager='test',
            active=1)

        self.safe_nova_client = clients.create_nova_client
        clients.create_nova_client = nova.fake_create_nova_client
        super(TestInstanceUpgrade, self).setUp()

    def tearDown(self):
        self.datastore.delete()
        self.datastore_version1.delete()
        self.datastore_version2.delete()
        clients.create_nova_client = self.safe_nova_client
        super(TestInstanceUpgrade, self).tearDown()

    @patch('trove.common.clients.create_neutron_client')
    @patch.object(task_api.API, 'get_client', Mock(return_value=Mock()))
    @patch.object(task_api.API, 'upgrade')
    @patch('trove.tests.fakes.nova.LOG')
    def test_upgrade(self, mock_logging, task_upgrade, mock_neutron_client):
        instance_model = DBInstance(
            InstanceTasks.NONE,
            id=str(uuid.uuid4()),
            name="TestUpgradeInstance",
            datastore_version_id=self.datastore_version1.id)
        instance_model.set_task_status(InstanceTasks.NONE)
        instance_model.save()
        instance_status = InstanceServiceStatus(
            ServiceStatuses.RUNNING,
            id=str(uuid.uuid4()),
            instance_id=instance_model.id)
        instance_status.save()
        self.assertIsNotNone(instance_model)
        instance = models.load_instance(models.Instance, self.context,
                                        instance_model.id)

        try:
            instance.upgrade(self.datastore_version2)

            self.assertEqual(self.datastore_version2.id,
                             instance.db_info.datastore_version_id)
            self.assertEqual(InstanceTasks.UPGRADING,
                             instance.db_info.task_status)
            self.assertTrue(task_upgrade.called)
        finally:
            instance_status.delete()
            instance_model.delete()


class TestReplication(trove_testtools.TestCase):

    def setUp(self):
        util.init_db()

        self.datastore = datastore_models.DBDatastore.create(
            id=str(uuid.uuid4()),
            name='name' + str(uuid.uuid4()),
            default_version_id=str(uuid.uuid4()))

        self.datastore_version = datastore_models.DBDatastoreVersion.create(
            id=self.datastore.default_version_id,
            name='name' + str(uuid.uuid4()),
            image_id=str(uuid.uuid4()),
            packages=str(uuid.uuid4()),
            datastore_id=self.datastore.id,
            manager='mysql',
            active=1)

        self.databases = []

        self.users = []

        self.master = DBInstance(
            InstanceTasks.NONE,
            id=str(uuid.uuid4()),
            name="TestMasterInstance",
            datastore_version_id=self.datastore_version.id,
            flavor_id=str(uuid.uuid4()),
            volume_size=2)
        self.master.set_task_status(InstanceTasks.NONE)
        self.master.save()
        self.master_status = InstanceServiceStatus(
            ServiceStatuses.RUNNING,
            id=str(uuid.uuid4()),
            instance_id=self.master.id)
        self.master_status.save()

        self.safe_nova_client = clients.create_nova_client
        clients.create_nova_client = nova.fake_create_nova_client

        self.swift_verify_patch = patch.object(models.Backup,
                                               'verify_swift_auth_token')
        self.addCleanup(self.swift_verify_patch.stop)
        self.swift_verify_patch.start()

        super(TestReplication, self).setUp()

    def tearDown(self):
        self.master.delete()
        self.master_status.delete()
        self.datastore.delete()
        self.datastore_version.delete()
        clients.create_nova_client = self.safe_nova_client
        super(TestReplication, self).tearDown()

    @patch('trove.instance.models.LOG')
    def test_replica_with_invalid_slave_of_id(self, mock_logging):
        self.assertRaises(exception.NotFound,
                          Instance.create,
                          None, 'name', 1, "UUID", [], [], self.datastore,
                          self.datastore_version, 2,
                          None, slave_of_id=str(uuid.uuid4()))


def trivial_key_function(id):
    return id * id


class TestInstanceKeyCaching(trove_testtools.TestCase):

    def setUp(self):
        super(TestInstanceKeyCaching, self).setUp()

    def tearDown(self):
        super(TestInstanceKeyCaching, self).tearDown()

    def test_basic_caching(self):
        keycache = instance_encryption_key_cache(trivial_key_function, 5)
        self.assertEqual(keycache[5], 25)
        self.assertEqual(keycache[5], 25)
        self.assertEqual(keycache[25], 625)

    def test_caching(self):
        keyfn = Mock(return_value=123)
        keycache = instance_encryption_key_cache(keyfn, 5)
        self.assertEqual(keycache[5], 123)
        self.assertEqual(keyfn.call_count, 1)
        self.assertEqual(keycache[5], 123)
        self.assertEqual(keyfn.call_count, 1)
        self.assertEqual(keycache[6], 123)
        self.assertEqual(keyfn.call_count, 2)
        self.assertEqual(keycache[7], 123)
        self.assertEqual(keyfn.call_count, 3)
        self.assertEqual(keycache[8], 123)
        self.assertEqual(keyfn.call_count, 4)
        self.assertEqual(keycache[9], 123)
        self.assertEqual(keyfn.call_count, 5)
        self.assertEqual(keycache[10], 123)
        self.assertEqual(keyfn.call_count, 6)
        self.assertEqual(keycache[10], 123)
        self.assertEqual(keyfn.call_count, 6)
        self.assertEqual(keycache[5], 123)
        self.assertEqual(keyfn.call_count, 7)

    # BUG(1650518): Cleanup in the Pike release
    def test_not_caching_none(self):
        keyfn = Mock(return_value=None)
        keycache = instance_encryption_key_cache(keyfn, 5)
        self.assertIsNone(keycache[30])
        self.assertEqual(keyfn.call_count, 1)
        self.assertIsNone(keycache[30])
        self.assertEqual(keyfn.call_count, 2)



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from nose.tools import assert_equal
from proboscis.asserts import assert_raises
from proboscis.asserts import assert_true
from proboscis import before_class
from proboscis import test
from troveclient.compat import exceptions

from trove import tests
from trove.tests.util.check import TypeCheck
from trove.tests.util import create_dbaas_client
from trove.tests.util import test_config
from trove.tests.util.users import Requirements

NAME = "nonexistent"


@test(groups=[tests.DBAAS_API_DATASTORES],
      depends_on_groups=[tests.DBAAS_API_VERSIONS])
class Datastores(object):
    @before_class
    def setUp(self):
        rd_user = test_config.users.find_user(
            Requirements(is_admin=False, services=["trove"]))
        rd_admin = test_config.users.find_user(
            Requirements(is_admin=True, services=["trove"]))
        self.rd_client = create_dbaas_client(rd_user)
        self.rd_admin = create_dbaas_client(rd_admin)

    @test
    def test_datastore_list_attrs(self):
        datastores = self.rd_client.datastores.list()
        for datastore in datastores:
            with TypeCheck('Datastore', datastore) as check:
                check.has_field("id", str)
                check.has_field("name", str)
                check.has_field("links", list)
                check.has_field("versions", list)

    @test
    def test_datastore_get(self):
        # Test get by name
        datastore_by_name = self.rd_client.datastores.get(
            test_config.dbaas_datastore)
        with TypeCheck('Datastore', datastore_by_name) as check:
            check.has_field("id", str)
            check.has_field("name", str)
            check.has_field("links", list)
        assert_equal(datastore_by_name.name, test_config.dbaas_datastore)

        # test get by id
        datastore_by_id = self.rd_client.datastores.get(
            datastore_by_name.id)
        with TypeCheck('Datastore', datastore_by_id) as check:
            check.has_field("id", str)
            check.has_field("name", str)
            check.has_field("links", list)
            check.has_field("versions", list)
        assert_equal(datastore_by_id.id, datastore_by_name.id)

    @test
    def test_datastore_not_found(self):
        try:
            assert_raises(exceptions.NotFound,
                          self.rd_client.datastores.get, NAME)
        except exceptions.BadRequest as e:
            assert_equal(e.message,
                         "Datastore '%s' cannot be found." % NAME)

    @test
    def test_create_inactive_datastore_by_admin(self):
        datastores = self.rd_admin.datastores.list()
        for ds in datastores:
            if ds.name == test_config.dbaas_datastore_name_no_versions:
                for version in ds.versions:
                    if version['name'] == 'inactive_version':
                        return

        # Create datastore version for testing
        # 'Test_Datastore_1' is also used in other test cases.
        # Will be deleted in test_delete_datastore_version
        self.rd_admin.mgmt_datastore_versions.create(
            "inactive_version", test_config.dbaas_datastore_name_no_versions,
            "test_manager", None, image_tags=['trove'],
            active='false', default='false'
        )

    @test(depends_on=[test_create_inactive_datastore_by_admin])
    def test_datastore_with_no_active_versions_is_hidden(self):
        datastores = self.rd_client.datastores.list()
        name_list = [datastore.name for datastore in datastores]

        assert_true(
            test_config.dbaas_datastore_name_no_versions not in name_list)

    @test(depends_on=[test_create_inactive_datastore_by_admin])
    def test_datastore_with_no_active_versions_is_visible_for_admin(self):
        datastores = self.rd_admin.datastores.list()
        name_list = [datastore.name for datastore in datastores]
        assert_true(test_config.dbaas_datastore_name_no_versions in name_list)


@test(groups=[tests.DBAAS_API_DATASTORES])
class DatastoreVersions(object):
    @before_class
    def setUp(self):
        rd_user = test_config.users.find_user(
            Requirements(is_admin=False, services=["trove"]))
        self.rd_client = create_dbaas_client(rd_user)
        self.datastore_active = self.rd_client.datastores.get(
            test_config.dbaas_datastore)
        self.datastore_version_active = self.rd_client.datastore_versions.list(
            self.datastore_active.id)[0]

    @test
    def test_datastore_version_list_attrs(self):
        versions = self.rd_client.datastore_versions.list(
            self.datastore_active.name)
        for version in versions:
            with TypeCheck('DatastoreVersion', version) as check:
                check.has_field("id", str)
                check.has_field("name", str)
                check.has_field("links", list)

    @test
    def test_datastore_version_get_attrs(self):
        version = self.rd_client.datastore_versions.get(
            self.datastore_active.name, self.datastore_version_active.name)
        with TypeCheck('DatastoreVersion', version) as check:
            check.has_field("id", str)
            check.has_field("name", str)
            check.has_field("datastore", str)
            check.has_field("links", list)
        assert_equal(version.name, self.datastore_version_active.name)

    @test
    def test_datastore_version_get_by_uuid_attrs(self):
        version = self.rd_client.datastore_versions.get_by_uuid(
            self.datastore_version_active.id)
        with TypeCheck('DatastoreVersion', version) as check:
            check.has_field("id", str)
            check.has_field("name", str)
            check.has_field("datastore", str)
            check.has_field("links", list)
        assert_equal(version.name, self.datastore_version_active.name)

    @test
    def test_datastore_version_not_found(self):
        assert_raises(exceptions.BadRequest,
                      self.rd_client.datastore_versions.get,
                      self.datastore_active.name, NAME)

    @test
    def test_datastore_version_list_by_uuid(self):
        versions = self.rd_client.datastore_versions.list(
            self.datastore_active.id)
        for version in versions:
            with TypeCheck('DatastoreVersion', version) as check:
                check.has_field("id", str)
                check.has_field("name", str)
                check.has_field("links", list)

    @test
    def test_datastore_version_get_by_uuid(self):
        version = self.rd_client.datastore_versions.get(
            self.datastore_active.id, self.datastore_version_active.id)
        with TypeCheck('DatastoreVersion', version) as check:
            check.has_field("id", str)
            check.has_field("name", str)
            check.has_field("datastore", str)
            check.has_field("links", list)
        assert_equal(version.name, self.datastore_version_active.name)

    @test
    def test_datastore_version_invalid_uuid(self):
        try:
            self.rd_client.datastore_versions.get_by_uuid(
                self.datastore_version_active.id)
        except exceptions.BadRequest as e:
            assert_equal(e.message,
                         "Datastore version '%s' cannot be found." %
                         test_config.dbaas_datastore_version)



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.tests.scenario.helpers.sql_helper import SqlHelper


class MysqlHelper(SqlHelper):

    def __init__(self, expected_override_name, report):
        super(MysqlHelper, self).__init__(expected_override_name, report,
                                          'mysql+pymysql')

    def get_helper_credentials(self):
        return {'name': 'lite', 'password': 'litepass', 'database': 'firstdb'}

    def get_helper_credentials_root(self):
        return {'name': 'root', 'password': 'rootpass'}

    def get_valid_database_definitions(self):
        return [{'name': 'db1', 'character_set': 'latin2',
                 'collate': 'latin2_general_ci'},
                {'name': 'db2'}, {"name": 'db3'}]

    def get_valid_user_definitions(self):
        return [{'name': 'a_user1', 'password': 'password1', 'databases': [],
                 'host': '127.0.0.1'},
                {'name': 'a_user2', 'password': 'password1',
                 'databases': [{'name': 'db1'}], 'host': '0.0.0.0'},
                {'name': 'a_user3', 'password': 'password1',
                 'databases': [{'name': 'db1'}, {'name': 'db2'}]}]

    def get_dynamic_group(self):
        return {'key_buffer_size': 10485760,
                'join_buffer_size': 10485760}

    def get_non_dynamic_group(self):
        return {'innodb_buffer_pool_size': 10485760,
                'long_query_time': 59.1}

    def get_invalid_groups(self):
        return [{'key_buffer_size': -1}, {"join_buffer_size": 'string_value'}]

    def get_exposed_user_log_names(self):
        return ['general', 'slow_query']

    def get_unexposed_sys_log_names(self):
        return ['guest', 'error']



# Copyright 2016 Tesora, Inc.
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import oslo_messaging as messaging
from osprofiler import profiler

from trove.common.context import TroveContext


class TroveSerializer(messaging.Serializer):
    """The Trove serializer class that handles class inheritance and base
       serializers.
    """

    def __init__(self, base):
        self._base = base

    def _serialize_entity(self, context, entity):
        return entity

    def serialize_entity(self, context, entity):
        if self._base:
            entity = self._base.serialize_entity(context, entity)

        return self._serialize_entity(context, entity)

    def _deserialize_entity(self, context, entity):
        return entity

    def deserialize_entity(self, context, entity):
        entity = self._deserialize_entity(context, entity)

        if self._base:
            entity = self._base.deserialize_entity(context, entity)

        return entity

    def _serialize_context(self, context):
        return context

    def serialize_context(self, context):
        if self._base:
            context = self._base.serialize_context(context)

        return self._serialize_context(context)

    def _deserialize_context(self, context):
        return context

    def deserialize_context(self, context):
        context = self._deserialize_context(context)

        if self._base:
            context = self._base.deserialize_context(context)

        return context


class TroveRequestContextSerializer(TroveSerializer):
    def _serialize_context(self, context):
        _context = context.to_dict()
        prof = profiler.get()
        if prof:
            trace_info = {
                "hmac_key": prof.hmac_key,
                "base_id": prof.get_base_id(),
                "parent_id": prof.get_id()
            }
            _context.update({"trace_info": trace_info})
        return _context

    def _deserialize_context(self, context):
        trace_info = context.pop("trace_info", None)
        if trace_info:
            profiler.init(**trace_info)
        return TroveContext.from_dict(context)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from trove.common.auth import admin_context
from trove.common import exception
from trove.common.i18n import _
from trove.common import wsgi
from trove.extensions.mgmt.quota import views
from trove.quota.models import Quota
from trove.quota.quota import QUOTAS as quota_engine

LOG = logging.getLogger(__name__)


class QuotaController(wsgi.Controller):
    """Controller for quota functionality."""

    def show(self, req, tenant_id, id):
        """Return all quotas for this tenant.

        Regular tenant can get his own resource quota.
        Admin user can get quota for any tenant.
        """
        LOG.info("Indexing quota info for tenant '%(id)s'\n"
                 "req : '%(req)s'\n\n", {"id": id, "req": req})

        context = req.environ[wsgi.CONTEXT_KEY]
        if id != tenant_id and not context.is_admin:
            raise exception.TroveOperationAuthError(
                tenant_id=tenant_id
            )

        usages = quota_engine.get_all_quota_usages_by_tenant(id)
        limits = quota_engine.get_all_quotas_by_tenant(id)
        for key in usages.keys():
            setattr(usages[key], "limit", limits[key].hard_limit)
        return wsgi.Result(views.QuotaUsageView(usages).data(), 200)

    @admin_context
    def update(self, req, body, tenant_id, id):
        LOG.info("Updating quota limits for tenant '%(id)s'\n"
                 "req : '%(req)s'\n\n", {"id": id, "req": req})

        if not body:
            raise exception.BadRequest(_("Invalid request body."))

        quotas = {}
        quota = None
        registered_resources = quota_engine.resources
        for resource, limit in body['quotas'].items():
            if limit is None:
                continue
            elif limit < -1:
                raise exception.QuotaLimitTooSmall(limit=limit,
                                                   resource=resource)
            if resource == "xmlns":
                continue
            if resource not in registered_resources:
                raise exception.QuotaResourceUnknown(unknown=resource)
            try:
                quota = Quota.find_by(tenant_id=id, resource=resource)
                quota.hard_limit = limit
                quota.save()
            except exception.ModelNotFoundError:
                quota = Quota.create(tenant_id=id,
                                     resource=resource,
                                     hard_limit=limit)

            quotas[resource] = quota

        return wsgi.Result(views.QuotaView(quotas).data(), 200)



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Routes all the requests to the task manager.
"""

from oslo_log import log as logging
import oslo_messaging as messaging

from trove.common import cfg
from trove.common import exception
from trove.common.i18n import _
from trove.common.notification import NotificationCastWrapper
from trove.common.strategies.cluster import strategy
from trove.guestagent import models as agent_models
from trove import rpc


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class API(object):
    """API for interacting with the task manager.

    API version history:
        * 1.0 - Initial version.

    When updating this API, also update API_LATEST_VERSION
    """

    # API_LATEST_VERSION should bump the minor number each time
    # a method signature is added or changed
    API_LATEST_VERSION = '1.0'

    # API_BASE_VERSION should only change on major version upgrade
    API_BASE_VERSION = '1.0'

    VERSION_ALIASES = {
        'icehouse': '1.0',
        'juno': '1.0',
        'kilo': '1.0',
        'liberty': '1.0',
        'mitaka': '1.0',
        'newton': '1.0',

        'latest': API_LATEST_VERSION
    }

    def __init__(self, context):
        self.context = context
        super(API, self).__init__()

        version_cap = self.VERSION_ALIASES.get(
            CONF.upgrade_levels.taskmanager, CONF.upgrade_levels.taskmanager)
        target = messaging.Target(topic=CONF.taskmanager_queue,
                                  version=version_cap)

        self.client = self.get_client(target, version_cap)

    def _cast(self, method_name, version, **kwargs):
        LOG.debug("Casting %s", method_name)
        with NotificationCastWrapper(self.context, 'taskmanager'):
            cctxt = self.client.prepare(version=version)
            cctxt.cast(self.context, method_name, **kwargs)

    def get_client(self, target, version_cap, serializer=None):
        if CONF.enable_secure_rpc_messaging:
            key = CONF.taskmanager_rpc_encr_key
        else:
            key = None

        return rpc.get_client(target, key=key,
                              version_cap=version_cap,
                              serializer=serializer)

    def _transform_obj(self, obj_ref):
        # Turn the object into a dictionary and remove the mgr
        if "__dict__" in dir(obj_ref):
            obj_dict = obj_ref.__dict__
            # We assume manager contains a object due to the *clients
            if obj_dict.get('manager'):
                del obj_dict['manager']
            return obj_dict
        raise ValueError(_("Could not transform %s") % obj_ref)

    def _delete_heartbeat(self, instance_id):
        agent_heart_beat = agent_models.AgentHeartBeat()
        try:
            heartbeat = agent_heart_beat.find_by_instance_id(instance_id)
            heartbeat.delete()
        except exception.ModelNotFoundError as e:
            LOG.error(e.message)

    def resize_volume(self, new_size, instance_id):
        LOG.debug("Making async call to resize volume for instance: %s",
                  instance_id)
        version = self.API_BASE_VERSION

        self._cast("resize_volume", version=version,
                   new_size=new_size,
                   instance_id=instance_id)

    def resize_flavor(self, instance_id, old_flavor, new_flavor):
        LOG.debug("Making async call to resize flavor for instance: %s",
                  instance_id)
        version = self.API_BASE_VERSION

        self._cast("resize_flavor", version=version,
                   instance_id=instance_id,
                   old_flavor=self._transform_obj(old_flavor),
                   new_flavor=self._transform_obj(new_flavor))

    def reboot(self, instance_id):
        LOG.debug("Making async call to reboot instance: %s", instance_id)
        version = self.API_BASE_VERSION

        self._cast("reboot", version=version, instance_id=instance_id)

    def restart(self, instance_id):
        LOG.debug("Making async call to restart instance: %s", instance_id)
        version = self.API_BASE_VERSION

        self._cast("restart", version=version, instance_id=instance_id)

    def detach_replica(self, instance_id):
        LOG.debug("Making async call to detach replica: %s", instance_id)
        version = self.API_BASE_VERSION

        self._cast("detach_replica", version=version,
                   instance_id=instance_id)

    def promote_to_replica_source(self, instance_id):
        LOG.debug("Making async call to promote replica to source: %s",
                  instance_id)
        version = self.API_BASE_VERSION
        self._cast("promote_to_replica_source", version=version,
                   instance_id=instance_id)

    def eject_replica_source(self, instance_id):
        LOG.debug("Making async call to eject replica source: %s",
                  instance_id)
        version = self.API_BASE_VERSION
        self._cast("eject_replica_source", version=version,
                   instance_id=instance_id)

    def migrate(self, instance_id, host):
        LOG.debug("Making async call to migrate instance: %s", instance_id)
        version = self.API_BASE_VERSION

        self._cast("migrate", version=version,
                   instance_id=instance_id, host=host)

    def rebuild(self, instance_id, image_id):
        LOG.debug("Making async call to rebuild instance: %s", instance_id)
        version = self.API_BASE_VERSION

        self._cast("rebuild", version=version, instance_id=instance_id,
                   image_id=image_id)

    def delete_instance(self, instance_id):
        LOG.debug("Making async call to delete instance: %s", instance_id)
        version = self.API_BASE_VERSION

        self._cast("delete_instance", version=version,
                   instance_id=instance_id)

    def update_access(self, instance_id, access):
        LOG.debug(f"Making async call to update instance: {instance_id}")
        version = self.API_BASE_VERSION

        self._cast("update_access", version=version,
                   instance_id=instance_id, access=access)

    def create_backup(self, backup_info, instance_id):
        LOG.debug("Making async call to create a backup for instance: %s",
                  instance_id)
        version = self.API_BASE_VERSION

        self._cast("create_backup", version=version,
                   backup_info=backup_info,
                   instance_id=instance_id)

    def delete_backup(self, backup_id):
        LOG.debug("Making async call to delete backup: %s", backup_id)
        version = self.API_BASE_VERSION

        self._cast("delete_backup", version=version, backup_id=backup_id)

    def create_instance(self, instance_id, name, flavor,
                        image_id, databases, users, datastore_manager,
                        packages, volume_size, backup_id=None,
                        availability_zone=None, root_password=None,
                        nics=None, overrides=None, slave_of_id=None,
                        cluster_config=None, volume_type=None,
                        modules=None, locality=None, access=None,
                        ds_version=None):

        LOG.debug("Making async call to create instance %s ", instance_id)
        version = self.API_BASE_VERSION
        self._cast("create_instance", version=version,
                   instance_id=instance_id, name=name,
                   flavor=self._transform_obj(flavor),
                   image_id=image_id,
                   databases=databases,
                   users=users,
                   datastore_manager=datastore_manager,
                   packages=packages,
                   volume_size=volume_size,
                   backup_id=backup_id,
                   availability_zone=availability_zone,
                   root_password=root_password,
                   nics=nics,
                   overrides=overrides,
                   slave_of_id=slave_of_id,
                   cluster_config=cluster_config,
                   volume_type=volume_type,
                   modules=modules, locality=locality, access=access,
                   ds_version=ds_version)

    def create_cluster(self, cluster_id):
        LOG.debug("Making async call to create cluster %s ", cluster_id)
        version = self.API_BASE_VERSION

        self._cast("create_cluster", version=version, cluster_id=cluster_id)

    def grow_cluster(self, cluster_id, new_instance_ids):
        LOG.debug("Making async call to grow cluster %s ", cluster_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "grow_cluster",
                   cluster_id=cluster_id, new_instance_ids=new_instance_ids)

    def shrink_cluster(self, cluster_id, instance_ids):
        LOG.debug("Making async call to shrink cluster %s ", cluster_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "shrink_cluster",
                   cluster_id=cluster_id, instance_ids=instance_ids)

    def delete_cluster(self, cluster_id):
        LOG.debug("Making async call to delete cluster %s ", cluster_id)
        version = self.API_BASE_VERSION

        self._cast("delete_cluster", version=version, cluster_id=cluster_id)

    def upgrade(self, instance_id, datastore_version_id):
        LOG.debug("Making async call to upgrade guest to datastore "
                  "version %s ", datastore_version_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "upgrade", instance_id=instance_id,
                   datastore_version_id=datastore_version_id)

    def restart_cluster(self, cluster_id):
        LOG.debug("Making async call to restart cluster %s ", cluster_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "restart_cluster", cluster_id=cluster_id)

    def upgrade_cluster(self, cluster_id, datastore_version_id):
        LOG.debug("Making async call to upgrade guest to datastore "
                  "version %s ", datastore_version_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "upgrade_cluster", cluster_id=cluster_id,
                   datastore_version_id=datastore_version_id)

    def reapply_module(self, module_id, md5, include_clustered,
                       batch_size, batch_delay, force):
        LOG.debug("Making async call to reapply module %s", module_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "reapply_module",
                   module_id=module_id, md5=md5,
                   include_clustered=include_clustered,
                   batch_size=batch_size, batch_delay=batch_delay, force=force)


def load(context, manager=None):
    if manager:
        task_manager_api_class = (strategy.load_taskmanager_strategy(manager)
                                  .task_manager_api_class)
    else:
        task_manager_api_class = API
    return task_manager_api_class(context)
